from cfinterface.components.register import Register
from cfinterface.files.registerfile import RegisterFile
from inewave.newave.modelos.modif import (
    TURBMAXT,
    TURBMINT,
    USINA,
    VAZMAXT,
    VOLMIN,
    VOLMAX,
    NUMCNJ,
    NUMMAQ,
    VAZMIN,
    CFUGA,
    CMONT,
    VMAXT,
    VMINT,
    VMINP,
    VAZMINT,
)


from typing import Type, TypeVar, List, Optional, Union


class Modif(RegisterFile):
    """
    Armazena os dados de entrada do NEWAVE referentes às alterações nas
    configurações das usinas hidroelétricas.
    """

    T = TypeVar("T")

    REGISTERS = [
        USINA,
        VOLMIN,
        VOLMAX,
        NUMCNJ,
        NUMMAQ,
        VAZMIN,
        CFUGA,
        CMONT,
        VMAXT,
        VMINT,
        VMINP,
        VAZMINT,
    ]

    def __init__(self, data=...) -> None:
        super().__init__(data)

    @classmethod
    def le_arquivo(cls, diretorio: str, nome_arquivo="modif.dat") -> "Modif":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="modif.dat"):
        self.write(diretorio, nome_arquivo)

    def __registros_por_tipo(self, registro: Type[T]) -> List[T]:
        """
        Obtém um gerador de blocos de um tipo, se houver algum no arquivo.

        :param bloco: Um tipo de bloco para ser lido
        :type bloco: T
        :param indice: O índice do bloco a ser acessado, dentre os do tipo
        :type indice: int

        """
        return [b for b in self.data.of_type(registro)]

    def __obtem_registros(self, tipo: Type[T]) -> List[T]:
        return self.__registros_por_tipo(tipo)

    def __obtem_registros_com_filtros(
        self, tipo_registro: Type[T], **kwargs
    ) -> Optional[Union[T, List[T]]]:
        def __atende(r) -> bool:
            condicoes: List[bool] = []
            for k, v in kwargs.items():
                if v is not None:
                    condicoes.append(getattr(r, k) == v)
            return all(condicoes)

        regs_filtro = [
            r for r in self.__obtem_registros(tipo_registro) if __atende(r)
        ]
        if len(regs_filtro) == 0:
            return None
        elif len(regs_filtro) == 1:
            return regs_filtro[0]
        else:
            return regs_filtro

    def cria_registro(self, anterior: Register, registro: Register):
        """
        Adiciona um registro ao arquivo após um outro registro previamente
        existente.
        Este método existe para retrocompatibilidade e deve ser substituído
        quando for suportado na classe :class:`RegisterFile`.
        """
        self.data.add_after(anterior, registro)

    def deleta_registro(self, registro: Register):
        """
        Remove um registro existente no arquivo.
        Este método existe para retrocompatibilidade e deve ser substituído
        quando for suportado na classe :class:`RegisterFile`.
        """
        self.data.remove(registro)

    def lista_registros(self, tipo: Type[T]) -> List[T]:
        """
        Lista todos os registros presentes no arquivo que tenham o tipo `T`.
        Este método existe para retrocompatibilidade e deve ser substituído
        quando for suportado na classe :class:`RegisterFile`.
        """
        return [r for r in self.data.of_type(tipo)]

    def append_registro(self, registro: Register):
        """
        Adiciona um registro ao arquivo na última posição.
        Este método existe para retrocompatibilidade e deve ser substituído
        quando for suportado na classe :class:`RegisterFile`.
        """
        self.data.append(registro)

    def preppend_registro(self, registro: Register):
        """
        Adiciona um registro ao arquivo na primeira posição.
        Este método existe para retrocompatibilidade e deve ser substituído
        quando for suportado na classe :class:`RegisterFile`.
        """
        self.data.preppend(registro)

    def usina(
        self, codigo: Optional[int] = None, nome: Optional[str] = None
    ) -> Optional[Union[USINA, List[USINA]]]:
        """
        Obtém um registro que define a usina modificada.

        :param codigo: código da usina modificada
        :type codigo: int | None
        :param nome: nome da usina
        :type nome: str | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`USINA` | list[:class:`USINA`] | None
        """
        return self.__obtem_registros_com_filtros(
            USINA, codigo=codigo, nome=nome
        )

    def volmin(
        self, volume: Optional[float] = None, unidade: Optional[str] = None
    ) -> Optional[Union[VOLMIN, List[VOLMIN]]]:
        """
        Obtém um registro que define um volume mínimo.

        :param volume: volume mínimo
        :type volume: float | None
        :param unidade: unidade do volume
        :type unidade: str | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`VOLMIN` | list[:class:`VOLMIN`] | None
        """
        return self.__obtem_registros_com_filtros(
            VOLMIN, volume=volume, unidade=unidade
        )

    def volmax(
        self, volume: Optional[float] = None, unidade: Optional[str] = None
    ) -> Optional[Union[VOLMAX, List[VOLMAX]]]:
        """
        Obtém um registro que define um volume máximo.

        :param volume: volume máximo
        :type volume: float | None
        :param unidade: unidade do volume
        :type unidade: str | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`VOLMAX` | list[:class:`VOLMAX`] | None
        """
        return self.__obtem_registros_com_filtros(
            VOLMAX, volume=volume, unidade=unidade
        )

    def numcnj(
        self,
        numero: Optional[int] = None,
    ) -> Optional[Union[NUMCNJ, List[NUMCNJ]]]:
        """
        Obtém um registro que um número de conjuntos de máquinas.

        :param numero: o número
        :type numero: int | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`NUMCNJ` | list[:class:`NUMCNJ`] | None
        """
        return self.__obtem_registros_com_filtros(NUMCNJ, numero=numero)

    def nummaq(
        self,
        conjunto: Optional[int] = None,
        numero_maquinas: Optional[int] = None,
    ) -> Optional[Union[NUMMAQ, List[NUMMAQ]]]:
        """
        Obtém um registro que um número de máquinas por
            conjunto de máquinas.

        :param conjunto: o conjunto
        :type conjunto: int | None
        :param numero_maquinas: o número de máquinas
        :type numero_maquinas: int | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`NUMMAQ` | list[:class:`NUMMAQ`] | None
        """
        return self.__obtem_registros_com_filtros(
            NUMMAQ, conjunto=conjunto, numero_maquinas=numero_maquinas
        )

    def vazmin(
        self, vazao: Optional[float] = None
    ) -> Optional[Union[VAZMIN, List[VAZMIN]]]:
        """
        Obtém um registro que define uma vazão mínima.

        :param vazao: vazão mínima
        :type vazao: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`VAZMIN` | list[:class:`VAZMIN`] | None
        """
        return self.__obtem_registros_com_filtros(VAZMIN, vazao=vazao)

    def cfuga(
        self,
        mes: Optional[int] = None,
        ano: Optional[int] = None,
        nivel: Optional[float] = None,
    ) -> Optional[Union[CFUGA, List[CFUGA]]]:
        """
        Obtém um registro que define o nível do canal de fuga.

        :param mes: mês de validade do nível
        :type mes: int | None
        :param ano: ano de validade do nível
        :type ano: int | None
        :param nivel: o nível
        :type nivel: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`CFUGA` | list[:class:`CFUGA`] | None
        """
        return self.__obtem_registros_com_filtros(
            CFUGA, mes=mes, ano=ano, nivel=nivel
        )

    def cmont(
        self,
        mes: Optional[int] = None,
        ano: Optional[int] = None,
        nivel: Optional[float] = None,
    ) -> Optional[Union[CMONT, List[CMONT]]]:
        """
        Obtém um registro que define o nível do canal de montante.

        :param mes: mês de validade do nível
        :type mes: int | None
        :param ano: ano de validade do nível
        :type ano: int | None
        :param nivel: o nível
        :type nivel: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`CMONT` | list[:class:`CMONT`] | None
        """
        return self.__obtem_registros_com_filtros(
            CMONT, mes=mes, ano=ano, nivel=nivel
        )

    def vmaxt(
        self,
        mes: Optional[int] = None,
        ano: Optional[int] = None,
        volume: Optional[float] = None,
        unidade: Optional[str] = None,
    ) -> Optional[Union[VMAXT, List[VMAXT]]]:
        """
        Obtém um registro que define o volume máximo por período.

        :param mes: mês de validade do volume
        :type mes: int | None
        :param ano: ano de validade do volume
        :type ano: int | None
        :param volume: o volume
        :type volume: float | None
        :param unidade: a unidade do volume
        :type unidade: str | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`VMAXT` | list[:class:`VMAXT`] | None
        """
        return self.__obtem_registros_com_filtros(
            VMAXT, mes=mes, ano=ano, volume=volume, unidade=unidade
        )

    def vmint(
        self,
        mes: Optional[int] = None,
        ano: Optional[int] = None,
        volume: Optional[float] = None,
        unidade: Optional[str] = None,
    ) -> Optional[Union[VMINT, List[VMINT]]]:
        """
        Obtém um registro que define o volume mínimo por período.

        :param mes: mês de validade do volume
        :type mes: int | None
        :param ano: ano de validade do volume
        :type ano: int | None
        :param volume: o volume
        :type volume: float | None
        :param unidade: a unidade do volume
        :type unidade: str | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`VMINT` | list[:class:`VMINT`] | None
        """
        return self.__obtem_registros_com_filtros(
            VMINT, mes=mes, ano=ano, volume=volume, unidade=unidade
        )

    def vminp(
        self,
        mes: Optional[int] = None,
        ano: Optional[int] = None,
        volume: Optional[float] = None,
        unidade: Optional[str] = None,
    ) -> Optional[Union[VMINP, List[VMINP]]]:
        """
        Obtém um registro que define o volume mínimo para penalidade.

        :param mes: mês de validade do volume
        :type mes: int | None
        :param ano: ano de validade do volume
        :type ano: int | None
        :param volume: o volume
        :type volume: float | None
        :param unidade: a unidade do volume
        :type unidade: str | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`VMINP` | list[:class:`VMINP`] | None
        """
        return self.__obtem_registros_com_filtros(
            VMINP, mes=mes, ano=ano, volume=volume, unidade=unidade
        )

    def vazmint(
        self,
        mes: Optional[int] = None,
        ano: Optional[int] = None,
        vazao: Optional[float] = None,
    ) -> Optional[Union[VAZMINT, List[VAZMINT]]]:
        """
        Obtém um registro que define a vazão mínima por período.

        :param mes: mês de validade da vazão
        :type mes: int | None
        :param ano: ano de validade da vazão
        :type ano: int | None
        :param vazao: a vazão mínima
        :type vazao: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`VAZMINT` | list[:class:`VAZMINT`] | None
        """
        return self.__obtem_registros_com_filtros(
            VAZMINT, mes=mes, ano=ano, vazao=vazao
        )

    def vazmaxt(
        self,
        mes: Optional[int] = None,
        ano: Optional[int] = None,
        vazao: Optional[float] = None,
    ) -> Optional[Union[VAZMAXT, List[VAZMAXT]]]:
        """
        Obtém um registro que define a vazão máxima por período.

        :param mes: mês de validade da vazão
        :type mes: int | None
        :param ano: ano de validade da vazão
        :type ano: int | None
        :param vazao: a vazão máxima
        :type vazao: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`VAZMAXT` | list[:class:`VAZMAXT`] | None
        """
        return self.__obtem_registros_com_filtros(
            VAZMAXT, mes=mes, ano=ano, vazao=vazao
        )

    def turbmaxt(
        self,
        mes: Optional[int] = None,
        ano: Optional[int] = None,
        turbinamento: Optional[float] = None,
    ) -> Optional[Union[TURBMAXT, List[TURBMAXT]]]:
        """
        Obtém um registro que define o turbinamento máximo por período.

        :param mes: mês de validade do turbinamento
        :type mes: int | None
        :param ano: ano de validade do turbinamento
        :type ano: int | None
        :param vazao: o turbinamento máximo
        :type vazao: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`TURBMAXT` | list[:class:`TURBMAXT`] | None
        """
        return self.__obtem_registros_com_filtros(
            TURBMAXT, mes=mes, ano=ano, turbinamento=turbinamento
        )

    def turbmint(
        self,
        mes: Optional[int] = None,
        ano: Optional[int] = None,
        turbinamento: Optional[float] = None,
    ) -> Optional[Union[TURBMINT, List[TURBMINT]]]:
        """
        Obtém um registro que define o turbinamento mínimo por período.
        :param mes: mês de validade do turbinamento
        :type mes: int | None
        :param ano: ano de validade do turbinamento
        :type ano: int | None
        :param vazao: o turbinamento mínimo
        :type vazao: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`TURBMINT` | list[:class:`TURBMINT`] | None
        """
        return self.__obtem_registros_com_filtros(
            TURBMINT, mes=mes, ano=ano, turbinamento=turbinamento
        )

    def modificacoes_usina(self, codigo: int) -> Optional[List[Register]]:
        """
        Filtra os registros que são associados a uma usina específica.

        :param codigo: O código da usina
        :type codigo: int
        :return: Os registros que modificam a usina
        :rtype: List[Register]
        """
        usinas = self.usina()
        if usinas is None or isinstance(usinas, USINA):
            return None
        reg_usina: List[USINA] = list(
            filter(lambda r: r.codigo == codigo, usinas)
        )
        modificacoes_usina: List[Register] = []
        if len(reg_usina) > 0:
            r = reg_usina[0].next
            while not (isinstance(r, USINA) or r.is_last):
                modificacoes_usina.append(r)
                r = r.next

        return modificacoes_usina



from inewave.config import MAX_UTES

from cfinterface.components.section import Section
from cfinterface.components.line import Line
from cfinterface.components.field import Field
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.floatfield import FloatField
from typing import List, IO
import numpy as np  # type: ignore
import pandas as pd  # type: ignore


class BlocoUTEClasT(Section):
    """
    Bloco com informações de cargas adicionais por mês/ano
    e por subsistema.
    """

    FIM_BLOCO = " 9999"

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        campos_ute: List[Field] = [
            IntegerField(4, 1),
            LiteralField(12, 6),
            LiteralField(12, 19),
        ]
        campos_custos: List[Field] = [
            FloatField(8, 29 + 8 * i, 2) for i in range(5)
        ]
        self.__linha = Line(campos_ute + campos_custos)
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoUTEClasT):
            return False
        bloco: BlocoUTEClasT = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            cols = [f"Custo {i}" for i in range(1, 6)]
            df = pd.DataFrame(tabela, columns=cols)
            df["Número"] = numero_ute
            df["Nome"] = nome_ute
            df["Tipo Combustível"] = tipo_combustivel
            df = df[["Número", "Nome", "Tipo Combustível"] + cols]
            return df

        # Salta as linhas adicionais
        for _ in range(2):
            self.__cabecalhos.append(file.readline())

        # Variáveis auxiliares
        numero_ute: List[int] = []
        nome_ute: List[str] = []
        tipo_combustivel: List[str] = []
        tabela = np.zeros((MAX_UTES, 5))
        i = 0
        while True:
            linha = file.readline()
            # Confere se acabou
            if len(linha) < 3:
                break
            if BlocoUTEClasT.FIM_BLOCO in linha:
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            dados = self.__linha.read(linha)
            tabela[i, :] = dados[3:]
            numero_ute.append(dados[0])
            nome_ute.append(dados[1])
            tipo_combustivel.append(dados[2])
            i += 1

    # Override
    def write(self, file: IO):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError("Dados do c_adic.dat não foram lidos com sucesso")

        for _, linha in self.data.iterrows():
            linha_lida: pd.Series = linha
            file.write(self.__linha.write(linha_lida.tolist()))
        file.write(BlocoUTEClasT.FIM_BLOCO + "\n")


class BlocoModificacaoUTEClasT(Section):
    """
    Bloco de modificações das informações das
    usinas cadastradas no arquivo do NEWAVE `clast.dat`.
    """

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        campos_modificacao: List[Field] = [
            IntegerField(4, 1),
            FloatField(7, 8, 2),
            IntegerField(2, 17),
            IntegerField(4, 20),
            IntegerField(2, 26),
            IntegerField(4, 29),
            LiteralField(12, 35),
        ]
        self.__linha = Line(campos_modificacao)
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoModificacaoUTEClasT):
            return False
        bloco: BlocoModificacaoUTEClasT = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            df = pd.DataFrame(
                data={
                    "Número": numero_ute,
                    "Custo": custo,
                    "Mês Início": mes_ini,
                    "Ano Início": ano_ini,
                    "Mês Fim": mes_fim,
                    "Ano Fim": ano_fim,
                    "Nome": nomes,
                }
            )
            return df

        # Salta as linhas adicionais
        for _ in range(2):
            self.__cabecalhos.append(file.readline())

        # Variáveis auxiliares
        numero_ute: List[int] = []
        custo: List[float] = []
        mes_ini: List[int] = []
        ano_ini: List[int] = []
        mes_fim: List[int] = []
        ano_fim: List[int] = []
        nomes: List[str] = []
        while True:
            linha = file.readline()
            # Confere se acabou
            if len(linha) < 3:
                if len(numero_ute) > 0:
                    self.data = converte_tabela_em_df()
                break
            dados = self.__linha.read(linha)
            numero_ute.append(dados[0])
            custo.append(dados[1])
            mes_ini.append(dados[2])
            ano_ini.append(dados[3])
            mes_fim.append(dados[4])
            ano_fim.append(dados[5])
            nomes.append(dados[6])

    # Override
    def write(self, file: IO):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError("Dados do clast.dat não foram lidos com sucesso")

        for _, linha in self.data.iterrows():
            linha_lida: pd.Series = linha
            for i in [2, 3, 4, 5]:
                linha_lida[i] = (
                    None if np.isnan(linha_lida[i]) else int(linha_lida[i])
                )
            file.write(self.__linha.write(linha_lida.tolist()))



from inewave.nwlistop.modelos.arquivos.arquivosin import (
    ArquivoSIN,
)
from inewave.nwlistop.modelos.vevminsin import VevminAnos


class VevminSIN(ArquivoSIN):
    """
    Armazena os dados das saídas referentes às violações da meta
    de energia da vazão mínima para o SIN.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `vevminsin.out`

    """

    BLOCKS = [
        VevminAnos,
    ]

    @classmethod
    def le_arquivo(
        cls, diretorio: str, nome_arquivo="vevminsin.out"
    ) -> "VevminSIN":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="vevminsin.out"):
        self.write(diretorio, nome_arquivo)



from inewave.config import MAX_LAG_ADTERM, MAX_UTES

from cfinterface.components.section import Section
from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.floatfield import FloatField
from typing import List, IO
import numpy as np  # type: ignore
import pandas as pd  # type: ignore


class BlocoUTEAdTerm(Section):
    """
    Bloco com os despachos antecipados das UTEs por patamar.
    """

    FIM_BLOCO = " 9999"

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha_ute = Line(
            [IntegerField(4, 1), LiteralField(12, 7), IntegerField(1, 21)]
        )
        self.__linha_despachos = Line([])
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoUTEAdTerm):
            return False
        bloco: BlocoUTEAdTerm = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            cols = [f"Patamar {i}" for i in range(1, n_pat + 1)]
            df = pd.DataFrame(tabela, columns=cols)
            df["Código UTE"] = codigo_utes
            df["Nome UTE"] = nome_utes
            df["Lag"] = lag_utes
            df = df[["Código UTE", "Nome UTE", "Lag"] + cols]
            return df

        # Salta as linhas adicionais
        for _ in range(2):
            self.__cabecalhos.append(file.readline())

        # Obtem o número de patamares e constroi a linha de despachos
        n_pat = len(
            [
                s.strip()
                for s in self.__cabecalhos[1].split(" ")
                if "XXXXXXX.XX" in s
            ]
        )
        self.__linha_despachos = Line(
            [FloatField(12, 22 + 12 * i, 2) for i in range(n_pat)]
        )

        # Variáveis auxiliares
        codigo_utes: List[int] = []
        nome_utes: List[str] = []
        lag_utes: List[int] = []
        codigo_atual = 0
        nome_atual = ""
        lag_atual = 0
        tabela = np.zeros((MAX_UTES * MAX_LAG_ADTERM, n_pat))
        i = 0
        while True:
            linha = file.readline()
            # Confere se acabou
            if len(linha) < 3:
                break
            if BlocoUTEAdTerm.FIM_BLOCO in linha:
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            # Senão, confere se é uma linha de UTE ou despacho
            if len(linha[:5].strip()) > 0:
                codigo_atual, nome_atual, lag_atual = self.__linha_ute.read(
                    linha
                )
            else:
                tabela[i, :] = self.__linha_despachos.read(linha)
                codigo_utes.append(codigo_atual)
                nome_utes.append(nome_atual)
                lag_utes.append(lag_atual)
                i += 1

    # Override
    def write(self, file: IO):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError("Dados do adterm.dat não foram lidos com sucesso")

        ultima_ute = 0
        cols_despachos = [c for c in list(self.data.columns) if "Patamar" in c]
        for _, linha in self.data.iterrows():
            linha_lida: pd.Series = linha
            if linha_lida["Código UTE"] != ultima_ute:
                ultima_ute = linha_lida["Código UTE"]
                file.write(
                    self.__linha_ute.write(
                        linha_lida[["Código UTE", "Nome UTE", "Lag"]].tolist()
                    )
                )
            file.write(
                self.__linha_despachos.write(
                    linha_lida[cols_despachos].tolist()
                )
            )
        file.write(BlocoUTEAdTerm.FIM_BLOCO + "\n")



from cfinterface.components.section import Section
from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.floatfield import FloatField
from cfinterface.components.datetimefield import DatetimeField
from typing import List, IO, Optional
from datetime import datetime
import pandas as pd  # type: ignore


class BlocoUHEExph(Section):
    """
    Bloco com as expensões das UHEs por máquina.
    """

    FIM_BLOCO = "9999"

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha_uhe = Line(
            [
                IntegerField(4, 0),
                LiteralField(12, 5),
                DatetimeField(size=7, starting_position=18, format="%m/%Y"),
                IntegerField(2, 31),
                FloatField(7, 35, decimal_digits=1),
                DatetimeField(size=7, starting_position=44, format="%m/%Y"),
                FloatField(6, 52, decimal_digits=1),
                IntegerField(2, 60),
                IntegerField(1, 64),
            ]
        )
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoUHEExph):
            return False
        bloco: BlocoUHEExph = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            df = pd.DataFrame()
            df["Código UHE"] = codigo_uhes
            df["Nome UHE"] = nome_uhes
            df["Início Enchimento"] = inicio_enchimento
            df["Duração"] = duracao
            df["Volume Morto"] = volume_morto
            df["Data de Entrada"] = data_entrada
            df["Potência"] = potencia
            df["Máquina"] = maquina
            df["Conjunto"] = conjunto
            return df

        # Salta as linhas adicionais
        for _ in range(3):
            self.__cabecalhos.append(file.readline())

        # Variáveis auxiliares
        codigo_uhes: List[Optional[int]] = []
        nome_uhes: List[Optional[str]] = []
        inicio_enchimento: List[Optional[datetime]] = []
        duracao: List[Optional[int]] = []
        volume_morto: List[Optional[float]] = []
        data_entrada: List[Optional[datetime]] = []
        potencia: List[Optional[float]] = []
        maquina: List[Optional[int]] = []
        conjunto: List[Optional[int]] = []

        codigo_atual: Optional[int] = None
        nome_atual: Optional[str] = None
        inicio_enchimento_atual: Optional[datetime] = None
        duracao_atual: Optional[int] = None
        volume_morto_atual: Optional[float] = None
        data_entrada_atual: Optional[datetime] = None
        potencia_atual: Optional[float] = None
        maquina_atual: Optional[int] = None
        conjunto_atual: Optional[int] = None

        while True:
            linha = file.readline()
            # Confere se acabou
            if len(linha) < 3:
                if len(codigo_uhes) > 0:
                    self.data = converte_tabela_em_df()
                break
            if BlocoUHEExph.FIM_BLOCO in linha:
                codigo_atual = None
                nome_atual = None
                continue
            (
                codigo_atual,
                nome_atual,
                inicio_enchimento_atual,
                duracao_atual,
                volume_morto_atual,
                data_entrada_atual,
                potencia_atual,
                maquina_atual,
                conjunto_atual,
            ) = self.__linha_uhe.read(linha)
            codigo_uhes.append(
                codigo_atual if codigo_atual is not None else codigo_uhes[-1]
            )
            nome_uhes.append(
                nome_atual if nome_atual is not None else nome_uhes[-1]
            )
            inicio_enchimento.append(inicio_enchimento_atual)
            duracao.append(duracao_atual)
            volume_morto.append(volume_morto_atual)
            data_entrada.append(data_entrada_atual)
            potencia.append(potencia_atual)
            maquina.append(maquina_atual)
            conjunto.append(conjunto_atual)

    # Override
    def write(self, file: IO):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError("Dados do exph.dat não foram lidos com sucesso")

        ultima_uhe = 0
        for _, linha in self.data.iterrows():
            linha_lida: pd.Series = linha
            if linha_lida["Código UHE"] != ultima_uhe and ultima_uhe != 0:
                file.write(f"{BlocoUHEExph.FIM_BLOCO}\n")
            ultima_uhe = int(linha_lida["Código UHE"])
            dados_linha = linha_lida.tolist()
            # Corrige posições opcionais
            for i in [3, -1, -2]:
                dados_linha[i] = (
                    int(dados_linha[i])
                    if not pd.isnull(dados_linha[i])
                    else None
                )
            file.write(self.__linha_uhe.write(dados_linha))
        file.write(f"{BlocoUHEExph.FIM_BLOCO}\n")



from typing import Optional
from cfinterface.components.register import Register
from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.floatfield import FloatField
from cfinterface.components.datetimefield import DatetimeField
from datetime import datetime


class RegistroEolicaCadastro(Register):
    """
    Registro que contém um cadastro de usina eólica.
    """

    IDENTIFIER = "EOLICA-CADASTRO"
    IDENTIFIER_DIGITS = 15
    LINE = Line(
        [
            IntegerField(),
            LiteralField(),
            LiteralField(),
            IntegerField(),
        ],
        delimiter=";",
    )

    @property
    def codigo_eolica(self) -> Optional[int]:
        return self.data[0]

    @codigo_eolica.setter
    def codigo_eolica(self, c: int):
        self.data[0] = c

    @property
    def nome_eolica(self) -> Optional[str]:
        return self.data[1]

    @nome_eolica.setter
    def nome_eolica(self, n: str):
        self.data[1] = n

    @property
    def identificador_eolica(self) -> Optional[str]:
        return self.data[2]

    @identificador_eolica.setter
    def identificador_eolica(self, n: str):
        self.data[2] = n

    @property
    def quantidade_conjuntos(self) -> Optional[int]:
        return self.data[3]

    @quantidade_conjuntos.setter
    def quantidade_conjuntos(self, q: int):
        self.data[3] = q


class RegistroEolicaCadastroConjuntoAerogeradores(Register):
    """
    Registro que contém o cadastro de um conjunto de
    aerogeradores em uma usina eólica.
    """

    IDENTIFIER = "EOLICA-CADASTRO-CONJUNTO-AEROGERADORES"
    IDENTIFIER_DIGITS = 38
    LINE = Line(
        [
            IntegerField(),
            IntegerField(),
            LiteralField(),
            IntegerField(),
        ],
        delimiter=";",
    )

    @property
    def codigo_eolica(self) -> Optional[int]:
        return self.data[0]

    @codigo_eolica.setter
    def codigo_eolica(self, c: int):
        self.data[0] = c

    @property
    def indice_conjunto(self) -> Optional[int]:
        return self.data[1]

    @indice_conjunto.setter
    def indice_conjunto(self, c: int):
        self.data[1] = c

    @property
    def nome_conjunto(self) -> Optional[str]:
        return self.data[2]

    @nome_conjunto.setter
    def nome_conjunto(self, n: str):
        self.data[2] = n

    @property
    def quantidade_aerogeradores(self) -> Optional[int]:
        return self.data[3]

    @quantidade_aerogeradores.setter
    def quantidade_aerogeradores(self, c: int):
        self.data[3] = c


class RegistroEolicaCadastroAerogerador(Register):
    """
    Registro que contém o cadastro de um aerogerador
    pertencente a um conjunto de aerogeradores em uma
    usina eólica.
    """

    IDENTIFIER = "EOLICA-CADASTRO-AEROGERADOR"
    IDENTIFIER_DIGITS = 27
    LINE = Line(
        [
            IntegerField(),
            IntegerField(),
            FloatField(decimal_digits=3),
            FloatField(decimal_digits=3),
            FloatField(decimal_digits=3),
            FloatField(decimal_digits=3),
            FloatField(decimal_digits=3),
            FloatField(decimal_digits=3),
            FloatField(decimal_digits=3),
        ],
        delimiter=";",
    )

    @property
    def codigo_eolica(self) -> Optional[int]:
        return self.data[0]

    @codigo_eolica.setter
    def codigo_eolica(self, c: int):
        self.data[0] = c

    @property
    def indice_conjunto(self) -> Optional[int]:
        return self.data[1]

    @indice_conjunto.setter
    def indice_conjunto(self, c: int):
        self.data[1] = c

    @property
    def velocidade_cutin(self) -> Optional[float]:
        return self.data[2]

    @velocidade_cutin.setter
    def velocidade_cutin(self, v: float):
        self.data[2] = v

    @property
    def velocidade_nominal(self) -> Optional[float]:
        return self.data[3]

    @velocidade_nominal.setter
    def velocidade_nominal(self, v: float):
        self.data[3] = v

    @property
    def velocidade_cutout(self) -> Optional[float]:
        return self.data[4]

    @velocidade_cutout.setter
    def velocidade_cutout(self, v: float):
        self.data[4] = v

    @property
    def potencia_velocidade_cutin(self) -> Optional[float]:
        return self.data[5]

    @potencia_velocidade_cutin.setter
    def potencia_velocidade_cutin(self, p: float):
        self.data[5] = p

    @property
    def potencia_velocidade_nominal(self) -> Optional[float]:
        return self.data[6]

    @potencia_velocidade_nominal.setter
    def potencia_velocidade_nominal(self, p: float):
        self.data[6] = p

    @property
    def potencia_velocidade_cutout(self) -> Optional[float]:
        return self.data[7]

    @potencia_velocidade_cutout.setter
    def potencia_velocidade_cutout(self, p: float):
        self.data[7] = p

    @property
    def altura_torre(self) -> Optional[float]:
        return self.data[8]

    @altura_torre.setter
    def altura_torre(self, h: float):
        self.data[8] = h


class RegistroEolicaConjuntoAerogeradoresQuantidadeOperandoPeriodo(Register):
    """
    Registro que contém a informação de início e fim de
    operação de um conjunto de aerogeradores.
    """

    IDENTIFIER = "EOLICA-CONJUNTO-AEROGERADORES-QUANTIDADE-OPERANDO-PERIODO"
    IDENTIFIER_DIGITS = 57
    LINE = Line(
        [
            IntegerField(),
            IntegerField(),
            DatetimeField(size=7, format="%Y/%m"),
            DatetimeField(size=7, format="%Y/%m"),
            IntegerField(),
        ],
        delimiter=";",
    )

    @property
    def codigo_eolica(self) -> Optional[int]:
        return self.data[0]

    @codigo_eolica.setter
    def codigo_eolica(self, c: int):
        self.data[0] = c

    @property
    def indice_conjunto(self) -> Optional[int]:
        return self.data[1]

    @indice_conjunto.setter
    def indice_conjunto(self, c: int):
        self.data[1] = c

    @property
    def periodo_inicial(self) -> Optional[datetime]:
        return self.data[2]

    @periodo_inicial.setter
    def periodo_inicial(self, p: datetime):
        self.data[2] = p

    @property
    def periodo_final(self) -> Optional[datetime]:
        return self.data[3]

    @periodo_final.setter
    def periodo_final(self, p: datetime):
        self.data[3] = p

    @property
    def numero_aerogeradores(self) -> Optional[int]:
        return self.data[4]

    @numero_aerogeradores.setter
    def numero_aerogeradores(self, n: int):
        self.data[4] = n


class RegistroEolicaConjuntoAerogeradoresPotenciaEfetiva(Register):
    """
    Registro que contém a informação de início e fim de
    operação para uma certa potência efetiva de um conjunto
    de aerogeradores.
    """

    IDENTIFIER = "EOLICA-CONJUNTO-AEROGERADORES-POTENCIAEFETIVA-PERIODO"
    IDENTIFIER_DIGITS = 53
    LINE = Line(
        [
            IntegerField(),
            IntegerField(),
            DatetimeField(size=7, format="%Y/%m"),
            DatetimeField(size=7, format="%Y/%m"),
            FloatField(decimal_digits=3),
        ],
        delimiter=";",
    )

    @property
    def codigo_eolica(self) -> Optional[int]:
        return self.data[0]

    @codigo_eolica.setter
    def codigo_eolica(self, c: int):
        self.data[0] = c

    @property
    def indice_conjunto(self) -> Optional[int]:
        return self.data[1]

    @indice_conjunto.setter
    def indice_conjunto(self, c: int):
        self.data[1] = c

    @property
    def periodo_inicial(self) -> Optional[datetime]:
        return self.data[2]

    @periodo_inicial.setter
    def periodo_inicial(self, p: datetime):
        self.data[2] = p

    @property
    def periodo_final(self) -> Optional[datetime]:
        return self.data[3]

    @periodo_final.setter
    def periodo_final(self, p: datetime):
        self.data[3] = p

    @property
    def potencia_efetiva(self) -> Optional[float]:
        return self.data[4]

    @potencia_efetiva.setter
    def potencia_efetiva(self, n: float):
        self.data[4] = n



from cfinterface.components.section import Section
from cfinterface.components.line import Line
from cfinterface.components.literalfield import LiteralField
from typing import IO


class NomeCaso(Section):
    """
    Bloco com o nome do caso do arquivo de
    entrada do NEWAVE `caso.dat`.
    """

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha = Line([LiteralField(40, 0)])

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, NomeCaso):
            return False
        bloco: NomeCaso = o
        if not all(
            [
                isinstance(self.data, str),
                isinstance(o.data, str),
            ]
        ):
            return False
        else:
            return self.data == bloco.data

    # Override
    def read(self, file: IO):
        self.data = self.__linha.read(file.readline())[0]

    # Override
    def write(self, file: IO):
        file.write(self.__linha.write([self.data]))


class CaminhoGerenciadorProcessos(Section):
    """
    Bloco com o caminho do gerenciador de processos usado
    no NEWAVE, contido no arquivo `caso.dat`.
    """

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha = Line([LiteralField(120, 0)])

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, CaminhoGerenciadorProcessos):
            return False
        bloco: CaminhoGerenciadorProcessos = o
        if not all(
            [
                isinstance(self.data, str),
                isinstance(o.data, str),
            ]
        ):
            return False
        else:
            return self.data == bloco.data

    # Override
    def read(self, file: IO):
        self.data = self.__linha.read(file.readline())[0]

    # Override
    def write(self, file: IO):
        file.write(self.__linha.write([self.data]))



from inewave.newave.modelos.parpeol import BlocoSerieVentosUEE
from inewave.newave.modelos.parpeol import BlocoCorrelVentosUEE
from inewave.newave.modelos.parpeol import BlocoSerieRuidosUEE
from inewave.newave.modelos.parpeol import BlocoCorrelRuidosUEE
from inewave.newave.modelos.parpeol import BlocoCorrelEspacialAnualConfig
from inewave.newave.modelos.parpeol import BlocoCorrelEspacialMensalConfig

from cfinterface.components.block import Block
from cfinterface.files.blockfile import BlockFile
from typing import Type, TypeVar, Optional, Any, List
import pandas as pd  # type: ignore


class PARpeol(BlockFile):
    """
    Armazena os dados de saída do NEWAVE referentes às
    séries sintéticas de ventos geradas pelo PAR(p).

    Esta classe lida com informações de saída do NEWAVE e
    cujas saídas devem ser compatíveis com as observadas através
    do NWLISTOP.

    """

    T = TypeVar("T")

    BLOCKS = [
        BlocoSerieVentosUEE,
        BlocoCorrelVentosUEE,
        BlocoSerieRuidosUEE,
        BlocoCorrelRuidosUEE,
        BlocoCorrelEspacialAnualConfig,
        BlocoCorrelEspacialMensalConfig,
    ]

    def __init__(self, data=...) -> None:
        super().__init__(data)
        self.__series_ventos = None
        self.__correl_series_ventos = None
        self.__series_ruido = None
        self.__correl_series_ruido = None
        self.__correl_espacial_anual = None
        self.__correl_espacial_mensal = None

    @classmethod
    def le_arquivo(
        cls, diretorio: str, nome_arquivo="parpeol.dat"
    ) -> "PARpeol":
        return cls.read(diretorio, nome_arquivo)

    def __uees(self) -> Optional[List[str]]:
        """
        Retorna a lista das UEEs lidos do arquivo.

        :return: Os nomes das UEEs
        :rtype: List[str]
        """
        if self.series_ventos_uee is None:
            return None
        else:
            return self.series_ventos_uee["UEE"].unique().tolist()

    def __bloco_por_tipo(self, bloco: Type[T], indice: int) -> Optional[T]:
        """
        Obtém um gerador de blocos de um tipo, se houver algum no arquivo.

        :param bloco: Um tipo de bloco para ser lido
        :type bloco: T
        :param indice: O índice do bloco a ser acessado, dentre os do tipo
        :type indice: int
        :return: O gerador de blocos, se houver
        :rtype: Optional[Generator[T], None, None]
        """
        try:
            return next(
                b
                for i, b in enumerate(self.data.of_type(bloco))
                if i == indice
            )
        except StopIteration:
            return None

    def __concatena_dados(self, bloco: Type[Block]) -> Optional[Any]:
        """
        Obtém os dados de um bloco se este existir dentre os blocos do arquivo.

        :param bloco: O tipo do bloco cujos dados serão extraídos
        :type bloco: Type[T]
        :param indice: Qual dos blocos do tipo será acessado
        :type indice: int, optional
        :return: Os dados do bloco, se existirem
        :rtype: Any
        """
        dados = pd.DataFrame()
        for b in self.data.of_type(bloco):
            if dados.empty:
                dados = b.data
            else:
                dados = pd.concat([dados, b.data], ignore_index=True)
        if not dados.empty:
            return dados
        else:
            return None

    def __adiciona_coluna_uee(
        self, df: Optional[pd.DataFrame]
    ) -> Optional[pd.DataFrame]:
        """
        Adiciona uma coluna com as UEEs de cada amostra, assumindo
        a mesma ordem das séries de ventos.

        :param df: O DataFrame que irá receber as UEEs
        :type df: pd.DataFrame
        :return: O DataFrame com as UEEs
        :rtype: pd.DataFrame
        """
        if df is None:
            return None
        uees = self.__uees()
        if uees is None:
            return None
        linhas_por_uee = df.shape[0] / len(uees)
        if int(linhas_por_uee) != linhas_por_uee:
            raise ValueError(
                f"{df.shape[0]} linhas não podem ser "
                + f"divididas em {len(uees)} grupos"
            )
        cols = list(df.columns)
        col_uee: List[str] = []
        for uee in uees:
            col_uee += [uee] * int(linhas_por_uee)
        df["UEE"] = col_uee
        return df[["UEE"] + cols]

    def __adiciona_coluna_uee_com_estagios(
        self, df: pd.DataFrame
    ) -> pd.DataFrame:
        """
        Adiciona uma coluna com as UEEs de cada amostra e outra
        com o estágio de cada uma, assumindo
        a mesma ordem das séries de ventos.

        :param df: O DataFrame que irá receber as UEEs
        :type df: pd.DataFrame
        :return: O DataFrame com as UEEs
        :rtype: pd.DataFrame
        """
        uees = self.__uees()
        if uees is None:
            return None
        linhas_por_uee = df.shape[0] / len(uees)
        if int(linhas_por_uee) != linhas_por_uee:
            raise ValueError(
                f"{df.shape[0]} linhas não podem ser "
                + f"divididas em {len(uees)} grupos"
            )
        cols = list(df.columns)
        col_uee: List[str] = []
        col_estagio: List[int] = []
        for uee in uees:
            col_uee += [uee] * int(linhas_por_uee)
            col_estagio += list(range(1, int(linhas_por_uee) + 1))
        df["UEE"] = col_uee
        df["Estágio"] = col_estagio
        return df[["UEE", "Estágio"] + cols]

    def __adiciona_coluna_uee_corrigindo_pre_pos(
        self, df: Optional[pd.DataFrame]
    ) -> Optional[pd.DataFrame]:
        """
        Adiciona uma coluna com as UEEs de cada amostra e outra
        com o estágio de cada uma, assumindo
        a mesma ordem das séries de ventos, e corrigindo os valores
        dos anos se houve períodos PRE e POS.

        :param df: O DataFrame que irá receber as UEEs
        :type df: pd.DataFrame
        :return: O DataFrame com as UEEs
        :rtype: pd.DataFrame
        """

        def converte_vetor_anos(anos: List[str], n: int) -> List[int]:
            # Descobre os anos pré e pós estudo
            numero_anos_pre = len([p for p in anos if p == "PRE"]) // n
            numero_anos_pos = len([p for p in anos if p == "POS"]) // n
            anos_estudo = [int(p) for p in anos if p not in ["PRE", "POS"]]
            # Descobre o primeiro ano de estudo
            primeiro_ano_estudo = sorted(anos_estudo)[0]
            # Descobre o último ano de estudo
            ultimo_ano_estudo = sorted(anos_estudo)[-1]
            # Substitui os anos pré e pós pelos valores específicos
            indice_inicio_pos = anos.index("POS") if "POS" in anos else -1
            for a in range(numero_anos_pre):
                idx_i = n * a
                idx_f = idx_i + n
                ano = primeiro_ano_estudo - (numero_anos_pre - a)
                anos[idx_i:idx_f] = [str(ano)] * n
            for a in range(numero_anos_pos):
                idx_i = indice_inicio_pos + n * a
                idx_f = idx_i + n
                ano = ultimo_ano_estudo + a + 1
                anos[idx_i:idx_f] = [str(ano)] * n
            return [int(a) for a in anos]

        if df is None:
            return None
        uees = self.__uees()
        if uees is None:
            return None
        linhas_por_uee = df.shape[0] / len(uees)
        if int(linhas_por_uee) != linhas_por_uee:
            raise ValueError(
                f"{df.shape[0]} linhas não podem ser "
                + f"divididas em {len(uees)} grupos"
            )
        cols = list(df.columns)
        col_uee: List[str] = []
        for uee in uees:
            col_uee += [uee] * int(linhas_por_uee)
        df["UEE"] = col_uee
        uee0 = uees[0]
        ano0 = df["Ano"].unique().tolist()[0]
        filtro = (df["Ano"] == ano0) & (df["UEE"] == uee0)
        n_series = df.loc[filtro].shape[0]
        for i, uee in enumerate(uees):
            i_i = i * int(linhas_por_uee)
            i_f = i_i + int(linhas_por_uee) - 1
            df.loc[i_i:i_f, "Ano"] = converte_vetor_anos(
                df.loc[i_i:i_f, "Ano"].tolist(), n_series
            )
        return df[["UEE"] + cols]

    @property
    def series_ventos_uee(self) -> Optional[pd.DataFrame]:
        """
        A tabela de séries de ventos para todas as configurações
        e UEEs, no mesmo formato do arquivo `parpeol.dat`.

        - UEE (`str`)
        - Configuração (`int`)
        - Ano (`int`)
        - Janeiro (`float`)
        - Fevereiro (`float`)
        - ...
        - Dezembro (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame | None
        """
        if self.__series_ventos is None:
            self.__series_ventos = self.__concatena_dados(BlocoSerieVentosUEE)
        return self.__series_ventos

    @property
    def series_ruido_uee(self) -> Optional[pd.DataFrame]:
        """
        A tabela de séries de ruído para todos os UEEs,
        no mesmo formato do arquivo `parpeol.dat`.

        - UEE (`str`)
        - Ano (`int`)
        - Série (`int`)
        - Janeiro (`float`)
        - Fevereiro (`float`)
        - ...
        - Dezembro (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame | None
        """
        if self.__series_ruido is None:
            self.__series_ruido = self.__concatena_dados(BlocoSerieRuidosUEE)
            self.__series_ruido = (
                self.__adiciona_coluna_uee_corrigindo_pre_pos(
                    self.__series_ruido
                )
            )
        return self.__series_ruido

    @property
    def correlacao_series_ventos_uee(self) -> Optional[pd.DataFrame]:
        """
        A tabela de correlação das séries de ventos para
        todas as configurações vigentes e UEEs,
        no mesmo formato do arquivo `parpeol.dat`.

        - UEE (`str`)
        - Data (`date`)
        - Lag 1 (`float`)
        - Lag 2 (`float`)
        - ...
        - Lag 11 (`float`)

        :return: A tabela como um DataFrame
        :rtype: pd.DataFrame | None
        """
        if self.__correl_series_ventos is None:
            self.__correl_series_ventos = self.__concatena_dados(
                BlocoCorrelVentosUEE
            )
            self.__correl_series_ventos = self.__adiciona_coluna_uee(
                self.__correl_series_ventos
            )
        return self.__correl_series_ventos

    @property
    def correlacao_series_ruidos_uee(self) -> Optional[pd.DataFrame]:
        """
        A tabela de correlação das séries de ruídos para
        todas as configurações vigentes e UEEs,
        no mesmo formato do arquivo `parpeol.dat`.

        - UEE (`str`)
        - Data (`date`)
        - Lag 1 (`float`)
        - Lag 2 (`float`)
        - ...
        - Lag 11 (`float`)

        :return: A tabela como um DataFrame
        :rtype: pd.DataFrame | None
        """
        if self.__correl_series_ruido is None:
            self.__correl_series_ruido = self.__concatena_dados(
                BlocoCorrelRuidosUEE
            )
            self.__correl_series_ruido = self.__adiciona_coluna_uee(
                self.__correl_series_ruido
            )
        return self.__correl_series_ruido

    @property
    def correlacao_espacial_anual(self) -> Optional[pd.DataFrame]:
        """
        A tabela de correlação para todas as configurações
        e REEs / UEEs, no mesmo formato do arquivo `parpeol.dat`.

        - Configuração (`int`)
        - UEE (`str`)
        - <Nome do REE 1> (`str`)
        - <Nome do REE 2> (`str`)
        - ...
        - <Nome do REE N> (`str`)
        - <Nome da UEE 1> (`str`)
        - ...
        - <Nome da UEE M> (`str`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame | None
        """
        if self.__correl_espacial_anual is None:
            self.__correl_espacial_anual = self.__concatena_dados(
                BlocoCorrelEspacialAnualConfig
            )
        return self.__correl_espacial_anual

    @property
    def correlacao_espacial_mensal(self) -> Optional[pd.DataFrame]:
        """
        A tabela de correlação para todas as configurações
        e REEs / UEEs, no mesmo formato do arquivo `parpeol.dat`.

        - Configuração (`int`)
        - REE (`str`)
        - MES (`int`)
        - <Nome do REE 1> (`str`)
        - <Nome do REE 2> (`str`)
        - ...
        - <Nome do REE N> (`str`)
        - <Nome da UEE 1> (`str`)
        - ...
        - <Nome da UEE M> (`str`)

        :return: A tabela como um DataFrame.
        :rtype: Optional[pd.DataFrame]
        """
        if self.__correl_espacial_mensal is None:
            self.__correl_espacial_mensal = self.__concatena_dados(
                BlocoCorrelEspacialMensalConfig
            )
        return self.__correl_espacial_mensal



from inewave.config import MESES_DF

from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.floatfield import FloatField


from inewave.nwlistop.modelos.blocos.valoresserie import ValoresSerie  # type: ignore


class EafsAnos(ValoresSerie):
    """
    Bloco com as informações das tabelas de energias afluente bruta
    por série e por mês/ano de estudo.
    """

    HEADER_LINE = Line([IntegerField(4, 10)])
    DATA_LINE = Line(
        [IntegerField(4, 2)]  # type: ignore
        + [
            FloatField(8, 6 + 9 * i, 0) for i in range(len(MESES_DF) + 1)  # type: ignore
        ]
    )



from inewave.newave.modelos.caso import NomeCaso, CaminhoGerenciadorProcessos

from cfinterface.files.sectionfile import SectionFile
from typing import Type, TypeVar, Optional


class Caso(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes às
    configurações das usinas hidrelétricas.

    Esta classe lida com informações de entrada fornecidas ao NEWAVE e
    que podem ser modificadas através do arquivo `modif.dat`.

    """

    T = TypeVar("T")

    SECTIONS = [NomeCaso, CaminhoGerenciadorProcessos]

    def __init__(self, data=...) -> None:
        super().__init__(data)

    @classmethod
    def le_arquivo(cls, diretorio: str, nome_arquivo="caso.dat") -> "Caso":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="caso.dat"):
        self.write(diretorio, nome_arquivo)

    def __bloco_por_tipo(self, bloco: Type[T], indice: int) -> Optional[T]:
        """
        Obtém um gerador de blocos de um tipo, se houver algum no arquivo.

        :param bloco: Um tipo de bloco para ser lido
        :type bloco: T
        :param indice: O índice do bloco a ser acessado, dentre os do tipo
        :type indice: int
        :return: O gerador de blocos, se houver
        :rtype: Optional[Generator[T], None, None]
        """
        try:
            return next(
                b
                for i, b in enumerate(self.data.of_type(bloco))
                if i == indice
            )
        except StopIteration:
            return None

    @property
    def arquivos(self) -> Optional[str]:
        """
        Caminho para o arquivo `arquivos.dat` de entrada do NEWAVE.

        :return: O caminho para o arquivo
        :rtype: str | None
        """
        b = self.__bloco_por_tipo(NomeCaso, 0)
        if b is not None:
            return b.data
        return None

    @arquivos.setter
    def arquivos(self, a: str):
        b = self.__bloco_por_tipo(NomeCaso, 0)
        if b is not None:
            b.data = a

    @property
    def gerenciador_processos(self) -> Optional[str]:
        """
        Caminho para o gerenciador de processos do NEWAVE.

        :return: O caminho para o arquivo
        :rtype: str | None
        """
        b = self.__bloco_por_tipo(CaminhoGerenciadorProcessos, 0)
        if b is not None:
            return b.data
        return None

    @gerenciador_processos.setter
    def gerenciador_processos(self, a: str):
        b = self.__bloco_por_tipo(CaminhoGerenciadorProcessos, 0)
        if b is not None:
            b.data = a



from cfinterface.components.block import Block
from cfinterface.components.line import Line

from inewave.config import MESES_DF, MAX_PATAMARES, MAX_SERIES_SINTETICAS

from typing import IO, List
import pandas as pd  # type: ignore
import numpy as np  # type: ignore


class ValoresSeriePatamar(Block):
    """
    Bloco com a informaçao de uma tabela por submercado, com
    entradas por série e patamar.
    """

    BEGIN_PATTERN = "     ANO: "
    END_PATTERN = " MEDIA"
    HEADER_LINE = Line([])
    DATA_LINE = Line([])

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha_ano = self.__class__.HEADER_LINE
        self.__linha = self.__class__.DATA_LINE

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, ValoresSeriePatamar):
            return False
        bloco: ValoresSeriePatamar = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            cols = MESES_DF + ["Média"]
            df = pd.DataFrame(tabela, columns=["Série"] + cols)
            df["Ano"] = self.__ano
            df["Patamar"] = patamares
            df = df[["Ano", "Série", "Patamar"] + cols]
            df = df.astype({"Série": "int64", "Ano": "int64"})
            return df

        self.__ano = self.__linha_ano.read(file.readline())[0]
        file.readline()

        # Variáveis auxiliares
        self.__serie_atual = 0
        tabela = np.zeros(
            (MAX_PATAMARES * MAX_SERIES_SINTETICAS, len(MESES_DF) + 2)
        )
        patamares: List[str] = []
        i = 0
        while True:
            linha = file.readline()
            if self.ends(linha):
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            dados = self.__linha.read(linha)
            if dados[0] is not None:
                self.__serie_atual = dados[0]
            tabela[i, 0] = self.__serie_atual
            patamares.append(dados[1])
            tabela[i, 1:] = dados[2:]
            i += 1



from inewave.config import MESES_DF

from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.floatfield import FloatField


from inewave.nwlistop.modelos.blocos.valoresseriepatamar import ValoresSeriePatamar  # type: ignore


class IntercambioAnos(ValoresSeriePatamar):
    """
    Bloco com as informações das tabelas de intercâmbio.
    """

    HEADER_LINE = Line([IntegerField(4, 10)])
    DATA_LINE = Line(
        [  # type: ignore
            IntegerField(4, 2),
            LiteralField(5, 6),
        ]
        + [FloatField(8, 12 + 9 * i, 1) for i in range(len(MESES_DF) + 1)]  # type: ignore
    )



MockRegistroEolicaCadastro = "EOLICA-CADASTRO ; 1 ; NEInterior ;  ; 1"

MockRegistroEolicaCadastroAerogerador = (
    "EOLICA-CADASTRO-AEROGERADOR ; 1 ; 1 ; 0 ; 0 ; 0 ; 0 ; 6058.890 ; 0 ; 0\n"
)

MockRegistroEolicaCadastroConjuntoAerogeradores = (
    "EOLICA-CADASTRO-CONJUNTO-AEROGERADORES ; 1 ; 1 ; NEInterior_cj ; 1\n"
)

MockRegistroEolicaConjuntoAerogeradoresQuantidadeOperandoPeriodo = "EOLICA-CONJUNTO-AEROGERADORES-QUANTIDADE-OPERANDO-PERIODO; 1 ; 1 ;2021/01; 2030/12; 1\n"

MockRegistroEolicaConjuntoAerogeradoresPotenciaEfetiva = "EOLICA-CONJUNTO-AEROGERADORES-POTENCIAEFETIVA-PERIODO; 1 ; 1 ;2021/01; 2030/12; 6058.890\n"

MockEolicaCadastro = [
    "& EOLICA-CADASTRO ; Codigo ; Nome ; Identificador ; QuantidadeIConjunto\n",
    "EOLICA-CADASTRO ; 1 ; NEInterior ;  ; 1\n",
    "EOLICA-CADASTRO ; 2 ; NELitoral ;  ;  1\n",
    "EOLICA-CADASTRO ; 3 ; NEPE ;  ;  1\n",
    "EOLICA-CADASTRO ; 4 ; SULInterior ; ; 1\n",
    "EOLICA-CADASTRO ; 5 ; SULLitoral ; ; 1\n",
    "\n",
    "& EOLICA-CADASTRO-CONJUNTO-AEROGERADORES ; CodigoEolica ; IConjunto ; NomeConjunto ; QuantidadeIAerogerador\n",
    "EOLICA-CADASTRO-CONJUNTO-AEROGERADORES ; 1 ; 1 ; NEInterior_cj ; 1\n",
    "EOLICA-CADASTRO-CONJUNTO-AEROGERADORES ; 2 ; 1 ; NELitoral_cj ; 1\n",
    "EOLICA-CADASTRO-CONJUNTO-AEROGERADORES ; 3 ; 1 ; NEPE_cj ; 1\n",
    "EOLICA-CADASTRO-CONJUNTO-AEROGERADORES ; 4 ; 1 ; SULInterior_cj ; 1\n",
    "EOLICA-CADASTRO-CONJUNTO-AEROGERADORES ; 5 ; 1 ; SULLitoral_cj  ; 1\n",
    "\n",
    "& EOLICA-CADASTRO-AEROGERADOR ; CodigoEolica ; IConjunto ; VelocidadeCutIn ; VelocidadeNominal ; VelocidadeCutOut ; PotenciaVelocidadeCutIn ; PotenciaVelocidadeNominal ; PotenciaVelocidadeCutOut ; AlturaTorre\n",
    "EOLICA-CADASTRO-AEROGERADOR ; 1 ; 1 ; 0 ; 0 ; 0 ; 0 ; 6058.890 ; 0 ; 0\n",
    "EOLICA-CADASTRO-AEROGERADOR ; 2 ; 1 ; 0 ; 0 ; 0 ; 0 ; 7359.195 ; 0 ; 0\n",
    "EOLICA-CADASTRO-AEROGERADOR ; 3 ; 1 ; 0 ; 0 ; 0 ; 0 ;  635.615 ; 0 ; 0\n",
    "EOLICA-CADASTRO-AEROGERADOR ; 4 ; 1 ; 0 ; 0 ; 0 ; 0 ;  292.200 ; 0 ; 0\n",
    "EOLICA-CADASTRO-AEROGERADOR ; 5 ; 1 ; 0 ; 0 ; 0 ; 0 ; 1651.190 ; 0 ; 0\n",
    "\n",
    "& EOLICA-CONJUNTO-AEROGERADORES-QUANTIDADE-OPERANDO-PERIODO; CodEolica; IConjAero  ;PerIni ;PerFin ;NumAeroConj\n",
    "EOLICA-CONJUNTO-AEROGERADORES-QUANTIDADE-OPERANDO-PERIODO; 1 ; 1 ;2021/01; 2030/12; 1\n",
    "EOLICA-CONJUNTO-AEROGERADORES-QUANTIDADE-OPERANDO-PERIODO; 2 ; 1 ;2021/01; 2030/12; 1\n",
    "EOLICA-CONJUNTO-AEROGERADORES-QUANTIDADE-OPERANDO-PERIODO; 3 ; 1 ;2021/01; 2030/12; 1\n",
    "EOLICA-CONJUNTO-AEROGERADORES-QUANTIDADE-OPERANDO-PERIODO; 4 ; 1 ;2021/01; 2030/12; 1\n",
    "EOLICA-CONJUNTO-AEROGERADORES-QUANTIDADE-OPERANDO-PERIODO; 5 ; 1 ;2021/01; 2030/12; 1\n",
    "\n",
    "& EOLICA-CONJUNTO-AEROGERADORES-POTENCIAEFETIVA-PERIODO; CodEolica; IConjAero  ;PerIni ;PerFin ;PotEfet \n",
    "EOLICA-CONJUNTO-AEROGERADORES-POTENCIAEFETIVA-PERIODO; 1 ; 1 ;2021/01; 2030/12; 6058.890\n",
    "EOLICA-CONJUNTO-AEROGERADORES-POTENCIAEFETIVA-PERIODO; 2 ; 1 ;2021/01; 2030/12; 7359.195\n",
    "EOLICA-CONJUNTO-AEROGERADORES-POTENCIAEFETIVA-PERIODO; 3 ; 1 ;2021/01; 2030/12;  635.615\n",
    "EOLICA-CONJUNTO-AEROGERADORES-POTENCIAEFETIVA-PERIODO; 4 ; 1 ;2021/01; 2030/12;  292.200\n",
    "EOLICA-CONJUNTO-AEROGERADORES-POTENCIAEFETIVA-PERIODO; 5 ; 1 ;2021/01; 2030/12; 1651.190\n",
    "\n",
]



from inewave.nwlistop.modelos.geolsin import GEAnos

from inewave.nwlistop.modelos.arquivos.arquivosinpatamar import (
    ArquivoSINPatamar,
)


class GeolSIN(ArquivoSINPatamar):
    """
    Armazena os dados das saídas referentes à geração eólica total
    para o SIN.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `geolsin.out`.
    """

    BLOCKS = [
        GEAnos,
    ]

    @classmethod
    def le_arquivo(
        cls, diretorio: str, nome_arquivo="geolsin.out"
    ) -> "GeolSIN":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="geolsin.out"):
        self.write(diretorio, nome_arquivo)



from inewave.newave.modelos.vazpast import BlocoVazPast

from cfinterface.files.sectionfile import SectionFile
from typing import Type, TypeVar, Optional
import pandas as pd  # type: ignore


class VazPast(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes às
    vazões anteriores ao período de planejamento.

    Esta classe lida com informações de entrada fornecidas ao NEWAVE e
    que são usadas juntos das contidas no arquivo `vazoes.dat`.

    """

    T = TypeVar("T")

    SECTIONS = [BlocoVazPast]

    def __init__(self, data=...) -> None:
        super().__init__(data)

    @classmethod
    def le_arquivo(
        cls, diretorio: str, nome_arquivo="vazpast.dat"
    ) -> "VazPast":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="vazpast.dat"):
        self.write(diretorio, nome_arquivo)

    def __bloco_por_tipo(self, bloco: Type[T], indice: int) -> Optional[T]:
        """
        Obtém um gerador de blocos de um tipo, se houver algum no arquivo.

        :param bloco: Um tipo de bloco para ser lido
        :type bloco: T
        :param indice: O índice do bloco a ser acessado, dentre os do tipo
        :type indice: int
        :return: O gerador de blocos, se houver
        :rtype: Optional[Generator[T], None, None]
        """
        try:
            return next(
                b
                for i, b in enumerate(self.data.of_type(bloco))
                if i == indice
            )
        except StopIteration:
            return None

    @property
    def tendencia(self) -> Optional[pd.DataFrame]:
        """
        Tabela com a tendência hidrológica por REE.

        - Índice (`int`)
        - Usina (`str`)
        - Janeiro (`float`)
        - ...
        - Dezembro (`float`)

        :return: A tabela como um DataFrame
        :rtype: pd.DataFrame | None
        """
        b = self.__bloco_por_tipo(BlocoVazPast, 0)
        if b is not None:
            return b.data
        return None

    @tendencia.setter
    def tendencia(self, df: pd.DataFrame):
        b = self.__bloco_por_tipo(BlocoVazPast, 0)
        if b is not None:
            b.data = df



from inewave.nwlistop.modelos.blocos.valoresserie import (
    ValoresSerie,
)

from cfinterface.files.blockfile import BlockFile
import pandas as pd  # type: ignore
from typing import TypeVar, Optional


class ArquivoSIN(BlockFile):
    """
    Armazena os dados das saídas por submercado.
    """

    T = TypeVar("T")

    BLOCKS = [ValoresSerie]

    def __init__(self, data=...) -> None:
        super().__init__(data)
        self.__valores = None

    @classmethod
    def le_arquivo(
        cls, diretorio: str, nome_arquivo="arq.out"
    ) -> "ArquivoSIN":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="arq.out"):
        self.write(diretorio, nome_arquivo)

    def __monta_tabela(self) -> pd.DataFrame:
        df = None
        for b in self.data.of_type(ValoresSerie):
            dados = b.data
            if dados is None:
                continue
            elif df is None:
                df = b.data
            else:
                df = pd.concat([df, b.data], ignore_index=True)
        return df

    @property
    def valores(self) -> Optional[pd.DataFrame]:
        """
        Tabela com os valores por patamar, por série e
        por mês/ano de estudo.

        - Ano (`int`)
        - Série (`int`)
        - Patamar (`str`)
        - Janeiro (`float`)
        - ...
        - Dezembro (`float`)

        :return: A tabela dos valores por patamar.
        :rtype: pd.DataFrame | None
        """
        if self.__valores is None:
            self.__valores = self.__monta_tabela()
        return self.__valores



# Rotinas de testes associadas ao arquivo hist-ventos.csv do NEWAVE
from datetime import datetime
from inewave.newave.modelos.eolicahistorico import (
    RegistroEolicaHistoricoVentoHorizonte,
    RegistroEolicaHistoricoVento,
)

from inewave.newave.eolicahistorico import EolicaHistorico

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.eolicahistorico import (
    MockRegistroEolicaHistoricoHorizonte,
    MockRegistroEolicaHistorico,
    MockEolicaHistorico,
)


def test_registro_eolica_historico_horizonte_eolicahistorico():

    m: MagicMock = mock_open(
        read_data="".join(MockRegistroEolicaHistoricoHorizonte)
    )
    r = RegistroEolicaHistoricoVentoHorizonte()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [datetime(1979, 1, 1), datetime(2016, 1, 1)]
    assert r.data_inicial == datetime(1979, 1, 1)
    r.data_inicial = datetime(1980, 1, 1)
    assert r.data_final == datetime(2016, 1, 1)
    r.data_final = datetime(2018, 1, 1)


def test_registro_eolica_historico_eolicahistorico():

    m: MagicMock = mock_open(read_data="".join(MockRegistroEolicaHistorico))
    r = RegistroEolicaHistoricoVento()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, datetime(1979, 1, 1), datetime(1979, 2, 1), 3.43, 1.0]
    assert r.codigo_eolica == 1
    r.codigo_eolica = 2
    assert r.data_inicial == datetime(1979, 1, 1)
    r.data_inicial = datetime(1980, 1, 1)
    assert r.data_final == datetime(1979, 2, 1)
    r.data_final = datetime(1980, 2, 1)
    assert r.velocidade == 3.43
    r.velocidade = 5.0
    assert r.direcao == 1.0
    r.direcao = 0.0


def test_atributos_encontrados_eolicahistorico():
    m: MagicMock = mock_open(read_data="".join(MockEolicaHistorico))
    with patch("builtins.open", m):
        e = EolicaHistorico.le_arquivo("")
        assert len(e.eolica_historico_vento()) > 0


def test_eq_eolicahistorico():
    m: MagicMock = mock_open(read_data="".join(MockEolicaHistorico))
    with patch("builtins.open", m):
        cf1 = EolicaHistorico.le_arquivo("")
        cf2 = EolicaHistorico.le_arquivo("")
        assert cf1 == cf2


def test_neq_eolicahistorico():
    m: MagicMock = mock_open(read_data="".join(MockEolicaHistorico))
    with patch("builtins.open", m):
        cf1 = EolicaHistorico.le_arquivo("")
        cf2 = EolicaHistorico.le_arquivo("")
        cf2.deleta_registro(cf1.eolica_historico_vento()[0])
        assert cf1 != cf2


def test_leitura_escrita_eolicahistorico():
    m_leitura: MagicMock = mock_open(read_data="".join(MockEolicaHistorico))
    with patch("builtins.open", m_leitura):
        cf1 = EolicaHistorico.le_arquivo("")
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        cf1.escreve_arquivo("", "")
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(2, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        cf2 = EolicaHistorico.le_arquivo("")
        assert cf1 == cf2



from inewave.nwlistop.modelos.blocos.usina import Usina
from inewave.nwlistop.modelos.arquivos.arquivousinapatamar import (
    ArquivoUsinaPatamar,
)
from inewave.nwlistop.modelos.geol import GEAnos


class Geol(ArquivoUsinaPatamar):
    """
    Armazena os dados das saídas referentes à geração eólica total
    por patamar, por usina.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `geol00x.out`, onde x varia conforme o
    PEE em questão.

    """

    BLOCKS = [
        Usina,
        GEAnos,
    ]

    @classmethod
    def le_arquivo(cls, diretorio: str, nome_arquivo="geol001.out") -> "Geol":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="geol001.out"):
        self.write(diretorio, nome_arquivo)



from inewave.config import MESES_DF

from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.floatfield import FloatField

from inewave.nwlistop.modelos.blocos.valoresserie import ValoresSerie


class EarmAnos(ValoresSerie):
    """
    Bloco com as informações das tabelas de energia armazenada
    final em MWmes para o SIN.
    """

    HEADER_LINE = Line([IntegerField(4, 10)])
    DATA_LINE = Line(
        [IntegerField(4, 2)]  # type: ignore
        + [
            FloatField(9, 6 + 9 * i, 0) for i in range(len(MESES_DF) + 1)  # type: ignore
        ]
    )



from inewave.nwlistcf import Nwlistcf

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.nwlistcf import MockNwlistcf


def test_atributos_encontrados_nwlistcf():
    m: MagicMock = mock_open(read_data="".join(MockNwlistcf))
    with patch("builtins.open", m):
        n = Nwlistcf.le_arquivo("")
        assert n.cortes is not None


def test_atributos_nao_encontrados_nwlistcf():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        n = Nwlistcf.le_arquivo("")
        assert n.cortes is None


def test_eq_nwlistcf():
    m: MagicMock = mock_open(read_data="".join(MockNwlistcf))
    with patch("builtins.open", m):
        n1 = Nwlistcf.le_arquivo("")
        n2 = Nwlistcf.le_arquivo("")
        assert n1 == n2


# Não deve ter teste de diferença, visto que o atributo é
# implementado como Lazy Property.



from cfinterface.files.registerfile import RegisterFile
from inewave.newave.modelos.hidr import RegistroUHEHidr
from inewave.config import MESES_ABREV
import pandas as pd  # type: ignore


from typing import TypeVar, List, Optional


class Hidr(RegisterFile):
    """
    Armazena os dados de entrada do NEWAVE referentes ao cadastro das
    usinas hidroelétricas.
    """

    T = TypeVar("T")

    REGISTERS = [RegistroUHEHidr]
    STORAGE = "BINARY"

    def __init__(self, data=...) -> None:
        super().__init__(data)
        self.__df: Optional[pd.DataFrame] = None

    @classmethod
    def le_arquivo(cls, diretorio: str, nome_arquivo="hidr.dat") -> "Hidr":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="hidr.dat"):
        self.__atualiza_registros()
        self.write(diretorio, nome_arquivo)

    def __monta_df_de_registros(self) -> Optional[pd.DataFrame]:
        registros: List[RegistroUHEHidr] = [
            r for r in self.data.of_type(RegistroUHEHidr)
        ]
        if len(registros) == 0:
            return None
        df = pd.DataFrame(
            columns=[
                "Nome",
                "Posto",
                "Subsistema",
                "Empresa",
                "Jusante",
                "Desvio",
                "Volume Mínimo",
                "Volume Máximo",
                "Volume Vertedouro",
                "Volume Desvio",
                "Cota Mínima",
                "Cota Máxima",
                *[f"A{i} VC" for i in range(5)],
                *[f"A{i} CA" for i in range(5)],
                *[f"Evaporação {m}" for m in MESES_ABREV],
                "Num. Conjuntos Máquinas",
                *[f"Num. Máquinas Conjunto {i}" for i in range(1, 6)],
                *[f"PotEf Conjunto {i}" for i in range(1, 6)],
                *[f"HEf Conjunto {i}" for i in range(1, 6)],
                *[f"QEf Conjunto {i}" for i in range(1, 6)],
                "Produtibilidade Específica",
                "Perdas",
                "Número PJUS",
                *[f"A{i} PJUS1" for i in range(5)],
                *[f"A{i} PJUS2" for i in range(5)],
                *[f"A{i} PJUS3" for i in range(5)],
                *[f"A{i} PJUS4" for i in range(5)],
                *[f"A{i} PJUS5" for i in range(5)],
                *[f"A{i} PJUS6" for i in range(5)],
                *[f"REF PJUS {i}" for i in range(1, 7)],
                "Canal de Fuga Médio",
                "Influencia Vert. Cfuga",
                "Fator Carga Max.",
                "Fator Carga Min.",
                "Vazão Mínima Histórica",
                "Num. Unidades Base",
                "Tipo Turbina",
                "Representação Conjunto",
                "TEIF",
                "IP",
                "Tipo de Perda",
                "Data",
                "Observação",
                "Volume de Referência",
                "Tipo de Regulação",
            ]
        )
        for i, r in enumerate(registros):
            df.loc[i + 1] = [
                r.nome,
                r.posto,
                r.subsistema,
                r.empresa,
                r.jusante,
                r.desvio,
                r.volume_minimo,
                r.volume_maximo,
                r.volume_vertedouro,
                r.volume_desvio,
                r.cota_minima,
                r.cota_maxima,
                *(r.polinomio_volume_cota),
                *(r.polinomio_cota_area),
                *(r.evaporacao),
                r.numero_conjuntos_maquinas,
                *(r.numero_maquinas_conjunto),
                *(r.potef_conjunto),
                *(r.hef_conjunto),
                *(r.qef_conjunto),
                r.produtibilidade_especifica,
                r.perdas,
                r.numero_polinomios_jusante,
                *(r.polinomios_jusante),
                r.canal_fuga_medio,
                r.influencia_vertimento_canal_fuga,
                r.fator_carga_maximo,
                r.fator_carga_minimo,
                r.vazao_minima_historica,
                r.numero_unidades_base,
                r.tipo_turbina,
                r.representacao_conjunto,
                r.teif,
                r.ip,
                r.tipo_perda,
                r.data_referencia,
                r.observacao,
                r.volume_referencia,
                r.tipo_regulacao,
            ]
        df.index.name = "Código"

        df = df.astype(
            {
                "Nome": str,
                "Posto": int,
                "Subsistema": int,
                "Empresa": int,
                "Jusante": int,
                "Desvio": int,
                "Volume Mínimo": float,
                "Volume Máximo": float,
                "Volume Vertedouro": float,
                "Volume Desvio": float,
                "Cota Mínima": float,
                "Cota Máxima": float,
                "Produtibilidade Específica": float,
                "Perdas": float,
                "Canal de Fuga Médio": float,
                "Influencia Vert. Cfuga": int,
                "Fator Carga Max.": float,
                "Fator Carga Min.": float,
                "Vazão Mínima Histórica": int,
                "Num. Unidades Base": int,
                "Tipo Turbina": int,
                "Representação Conjunto": int,
                "TEIF": float,
                "IP": float,
                "Tipo de Perda": int,
                "Data": str,
                "Observação": str,
                "Volume de Referência": float,
                "Tipo de Regulação": str,
            },
        )
        return df

    def __atualiza_registros(self):
        registros: List[RegistroUHEHidr] = [r for r in self.data][1:]
        for (_, linha), r in zip(self.cadastro.iterrows(), registros):
            r.nome = linha["Nome"]
            r.posto = linha["Posto"]
            r.subsistema = linha["Subsistema"]
            r.empresa = linha["Empresa"]
            r.jusante = linha["Jusante"]
            r.desvio = linha["Desvio"]
            r.volume_minimo = linha["Volume Mínimo"]
            r.volume_maximo = linha["Volume Máximo"]
            r.volume_vertedouro = linha["Volume Vertedouro"]
            r.volume_desvio = linha["Volume Desvio"]
            r.cota_minima = linha["Cota Mínima"]
            r.cota_maxima = linha["Cota Máxima"]
            r.polinomio_volume_cota = linha[
                [f"A{i} VC" for i in range(5)]
            ].tolist()
            r.polinomio_cota_area = linha[
                [f"A{i} CA" for i in range(5)]
            ].tolist()
            r.evaporacao = linha[
                [f"Evaporação {m}" for m in MESES_ABREV]
            ].tolist()
            r.numero_conjuntos_maquinas = linha["Num. Conjuntos Máquinas"]
            r.potef_conjunto = linha[
                [f"PotEf Conjunto {i}" for i in range(1, 6)]
            ].tolist()
            r.hef_conjunto = linha[
                [f"HEf Conjunto {i}" for i in range(1, 6)]
            ].tolist()
            r.qef_conjunto = linha[
                [f"QEf Conjunto {i}" for i in range(1, 6)]
            ].tolist()
            r.produtibilidade_especifica = linha["Produtibilidade Específica"]
            r.perdas = linha["Perdas"]
            r.numero_polinomios_jusante = linha["Número PJUS"]
            r.polinomios_jusante = linha[
                [f"A{i} PJUS1" for i in range(5)]
                + [f"A{i} PJUS2" for i in range(5)]
                + [f"A{i} PJUS3" for i in range(5)]
                + [f"A{i} PJUS4" for i in range(5)]
                + [f"A{i} PJUS5" for i in range(5)]
                + [f"A{i} PJUS6" for i in range(5)]
                + [f"REF PJUS {i}" for i in range(1, 7)]
            ].tolist()
            r.canal_fuga_medio = linha["Canal de Fuga Médio"]
            r.influencia_vertimento_canal_fuga = linha[
                "Influencia Vert. Cfuga"
            ]
            r.fator_carga_maximo = linha["Fator Carga Max."]
            r.fator_carga_minimo = linha["Fator Carga Min."]
            r.vazao_minima_historica = linha["Vazão Mínima Histórica"]
            r.numero_unidades_base = linha["Num. Unidades Base"]
            r.tipo_turbina = linha["Tipo Turbina"]
            r.representacao_conjunto = linha["Representação Conjunto"]
            r.teif = linha["TEIF"]
            r.ip = linha["IP"]
            r.tipo_perda = linha["Tipo de Perda"]
            r.data_referencia = linha["Data"]
            r.observacao = linha["Observação"]
            r.volume_referencia = linha["Volume de Referência"]
            r.tipo_regulacao = linha["Tipo de Regulação"]

    @property
    def cadastro(self) -> pd.DataFrame:
        """
        Obtém a tabela com os dados cadastrais existentes no arquivo
        binário.

        - Nome (`str`): nome da usina (12 caracteres)
        - Posto (`int`): posto de vazão natural da usina
        - Subsistema (`int`): subsistema da usina
        - Empresa (`int`): agente responsável pela usina
        - Jusante (`int`): posto à jusante da usina
        - Desvio (`float`): TODO
        - Volume Mínimo (`float`): volume mínimo da usina (hm3)
        - Volume Máximo (`float`): volume máximo da usina (hm3)
        - Volume Vertedouro (`float`): volume do vertedouro da usina (hm3)
        - Volume Desvio (`float`): TODO
        - Cota Mínima (`float`): cota mínima da usina (m)
        - Cota Máxima (`float`): cota máxima da usina (m)
        - A[0-4] VC (`float`): coeficientes do polinômio volume-cota
        - A[0-4] CA (`float`): coeficientes do polinômio cota-área
        - Evaporação [JAN..DEZ] (`float`): coeficientes de evaporação (mm)
        - Num Conjunto Máquinas (`int`): número de conjuntos de máquinas
        - Num Máquinas Conjunto [1-5] (`int`): máquinas por conjunto
        - PotEf. Conjunto [1-5] (`float`): potência das máquinas (MWmed)
        - HEf Conjunto [1-5]: alturas nominais de queda por conjunto (m)
        - QEf Conjunto [1-5]: vazões nominais por conjunto (m3/s)
        - Produtibilidade Específica (`float`): produtibilidade específica
        - Perdas (`float`): perdas da usina
        - Número PJUS (`int`): número de polinômios de jusante
        - C[0-4] PJUS[1-6] (`float`): coeficientes de cada polinjus
        - REF PJUS [1-6] (`float`): coeficientes do polinjus de referência
        - Canal de Fuga Médio (`float`): cota média do canal de fuga (m)
        - Influencia Vert. Cfuga (`int`): TODO (0 ou 1)
        - Fator Carga Max. (`float`): TODO (%)
        - Fator Carga Min. (`float`): TODO (%)
        - Vazão Mínima Histórica (`float`): vazão mínima da usina (m3/s)
        - Num. Unidades Base (`int`): TODO
        - Tipo Turbina (`int`): TODO
        - Representação Conjunto (`int`): TODO
        - TEIF (`float`): TODO (%)
        - IP (`float`): TODO (%)
        - Tipo de Perda (`int`): TODO
        - Data (`str`): DD-MM-AA
        - Observação (`str`): observação qualquer sobre a usina
        - Volume de Referência (`float`): TODO (hm3)
        - Tipo de Regulação (`str`): D, S ou M

        :return: A tabela com os dados cadastrais
        :rtype: List[pd.DataFrame]
        """
        if self.__df is None:
            self.__df = self.__monta_df_de_registros()
        return self.__df

    @cadastro.setter
    def cadastro(self, df: pd.DataFrame):
        self.__df = df



from inewave.nwlistop.ghmaxr import Ghmaxr

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.ghmaxr import MockGhmaxr


def test_atributos_encontrados_ghmaxr():
    m: MagicMock = mock_open(read_data="".join(MockGhmaxr))
    with patch("builtins.open", m):
        n = Ghmaxr.le_arquivo("")
        assert n.ree is not None
        assert n.ree == "SUDESTE"
        assert n.valores is not None
        assert n.valores.iloc[0, 0] == 2021
        assert n.valores.iloc[-1, -1] == 7686.3


def test_atributos_nao_encontrados_ghmaxr():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        n = Ghmaxr.le_arquivo("")
        assert n.ree is None
        assert n.valores is None


def test_eq_ghmaxr():
    m: MagicMock = mock_open(read_data="".join(MockGhmaxr))
    with patch("builtins.open", m):
        n1 = Ghmaxr.le_arquivo("")
        n2 = Ghmaxr.le_arquivo("")
        assert n1 == n2


# Não deve ter teste de diferença, visto que o atributo é
# implementado como Lazy Property.



from inewave.nwlistop.modelos.blocos.ree import REE
from inewave.nwlistop.modelos.arquivos.arquivoree import ArquivoREE
from inewave.nwlistop.modelos.eaf import EafsAnos


class Eaf(ArquivoREE):
    """
    Armazena os dados das saídas referentes às energias
    afluentes, por REE.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `eaf00x.out`, onde x varia conforme o
    REE em questão.

    """

    BLOCKS = [
        REE,
        EafsAnos,
    ]

    @classmethod
    def le_arquivo(cls, diretorio: str, nome_arquivo="eaf001.out") -> "Eaf":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="eaf001.out"):
        self.write(diretorio, nome_arquivo)



# Inclui os membros

from .cmarg import Cmarg  # noqa
from .cmargmed import CmargMed  # noqa
from .coper import Coper  # noqa
from .cterm import Cterm  # noqa
from .ctermsin import CtermSIN  # noqa
from .dlppdfmax import DLPPdfmax  # noqa
from .dlppdfmaxs import DLPPdfmaxs  # noqa
from .dlpptbmax import DLPPtbmax  # noqa
from .dlpptbmaxm import DLPPtbmaxm  # noqa
from .dlpptbmaxs import DLPPtbmaxs  # noqa
from .eafb import Eafb  # noqa
from .eafbm import Eafbm  # noqa
from .eafbsin import EafbSIN  # noqa
from .earmf import Earmf  # noqa
from .earmfm import Earmfm  # noqa
from .earmfsin import EarmfSIN  # noqa
from .earmfp import Earmfp  # noqa
from .earmfpm import Earmfpm  # noqa
from .earmfpsin import EarmfpSIN  # noqa
from .geol import Geol  # noqa
from .geolm import Geolm  # noqa
from .geolsin import GeolSIN  # noqa
from .ghiduh import GhidUH  # noqa
from .ghtot import Ghtot  # noqa
from .ghtotm import Ghtotm  # noqa
from .ghtotsin import GhtotSIN  # noqa
from .ghmax import Ghmax  # noqa
from .ghmaxm import Ghmaxm  # noqa
from .ghmaxsin import GhmaxSIN  # noqa
from .ghmaxr import Ghmaxr  # noqa
from .ghmaxmr import Ghmaxmr  # noqa
from .ghmaxrsin import GhmaxrSIN  # noqa
from .gttot import Gttot  # noqa
from .gttotsin import GttotSIN  # noqa
from .intercambio import Intercambio  # noqa
from .invade import Invade  # noqa
from .invadem import Invadem  # noqa
from .mediasmerc import MediasMerc  # noqa
from .mediassin import MediasSIN  # noqa
from .qafluh import QaflUH  # noqa
from .qincruh import QincrUH  # noqa
from .rhslppdf import RHSLPPdf  # noqa
from .rhslpptb import RHSLPPtb  # noqa
from .varmuh import VarmUH  # noqa
from .varmpuh import VarmpUH  # noqa
from .vento import Vento  # noqa
from .vertuh import VertUH  # noqa
from .vturuh import VturUH  # noqa
from .verturb import Verturb  # noqa
from .verturbm import Verturbm  # noqa
from .verturbsin import VerturbSIN  # noqa
from .vevmin import Vevmin  # noqa
from .vevminm import Vevminm  # noqa
from .vevminsin import VevminSIN  # noqa



class ForwardH:
    """
    Armazena os dados de saída do NEWAVE referentes ao cabeçalho
    dos dados da simulação forward.

    **Parâmetros**

    """

    def __init__(self) -> None:
        pass



from inewave.newave.modelos.pmo import BlocoEafPastTendenciaHidrolPMO
from inewave.newave.modelos.pmo import BlocoEafPastCfugaMedioPMO
from inewave.newave.modelos.pmo import BlocoConvergenciaPMO
from inewave.newave.modelos.pmo import BlocoConfiguracoesExpansaoPMO
from inewave.newave.modelos.pmo import BlocoMARSPMO
from inewave.newave.modelos.pmo import BlocoRiscoDeficitENSPMO
from inewave.newave.modelos.pmo import BlocoCustoOperacaoPMO
from inewave.newave.modelos.pmo import BlocoCustoOperacaoTotalPMO

from cfinterface.components.block import Block
from cfinterface.files.blockfile import BlockFile
from typing import Type, TypeVar, Optional, Any
import pandas as pd  # type: ignore


class PMO(BlockFile):
    """
    Armazena os dados de entrada do NEWAVE referentes ao
    acompanhamento do programa.

    Esta classe lida com as informações de entrada fornecidas ao
    NEWAVE e reproduzidas no `pmo.dat`, bem como as saídas finais
    da execução: custos de operação, energias, déficit, etc.

    Em versões futuras, esta classe pode passar a ler os dados
    de execução intermediárias do programa.

    """

    T = TypeVar("T")

    BLOCKS = [
        BlocoEafPastTendenciaHidrolPMO,
        BlocoEafPastCfugaMedioPMO,
        BlocoConvergenciaPMO,
        BlocoConfiguracoesExpansaoPMO,
        BlocoMARSPMO,
        BlocoRiscoDeficitENSPMO,
        BlocoCustoOperacaoPMO,
        BlocoCustoOperacaoTotalPMO,
    ]

    @classmethod
    def le_arquivo(cls, diretorio: str, nome_arquivo="pmo.dat") -> "PMO":
        return cls.read(diretorio, nome_arquivo)

    def __bloco_por_tipo(self, bloco: Type[T], indice: int) -> Optional[T]:
        """
        Obtém um gerador de blocos de um tipo, se houver algum no arquivo.

        :param bloco: Um tipo de bloco para ser lido
        :type bloco: T
        :param indice: O índice do bloco a ser acessado, dentre os do tipo
        :type indice: int
        :return: O gerador de blocos, se houver
        :rtype: Optional[Generator[T], None, None]
        """
        try:
            return next(
                b
                for i, b in enumerate(self.data.of_type(bloco))
                if i == indice
            )
        except StopIteration:
            return None

    def __extrai_dados_se_existe(
        self, bloco: Type[Block], indice: int = 0
    ) -> Optional[Any]:
        """
        Obtém os dados de um bloco se este existir dentre os blocos do arquivo.

        :param bloco: O tipo do bloco cujos dados serão extraídos
        :type bloco: Type[T]
        :param indice: Qual dos blocos do tipo será acessado
        :type indice: int, optional
        :return: Os dados do bloco, se existirem
        :rtype: Any
        """
        b = self.__bloco_por_tipo(bloco, indice)
        if b is not None:
            return b.data
        return None

    @property
    def eafpast_tendencia_hidrologica(self) -> Optional[pd.DataFrame]:
        """
        Energias afluentes passadas por REE para análise da tendência
        hidrológica, em relação à primeira configuração do sistema,
        em MWmes.

        - REE (`str`)
        - Janeiro (`float`)
        - Fevereiro (`float`)
        - ...
        - Dezembro (`float`)

        :return: A tendência hidrológica em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        return self.__extrai_dados_se_existe(BlocoEafPastTendenciaHidrolPMO)

    @property
    def eafpast_cfuga_medio(self) -> Optional[pd.DataFrame]:
        """
        Energias afluentes passadas por REE considerando canal de
        fuga médio, em relação à primeira configuração do sistema,
        em MWmes.

        - REE (`str`)
        - Janeiro (`float`)
        - Fevereiro (`float`)
        - ...
        - Dezembro (`float`)

        :return: As energias afluentes passadas.
        :rtype: pd.DataFrame | None
        """
        return self.__extrai_dados_se_existe(BlocoEafPastCfugaMedioPMO)

    @property
    def configuracoes_entrada_reservatorio(self) -> Optional[pd.DataFrame]:
        """
        Configurações do sistema em cada período devido a entrada
        de reservatórios e/ou potência de base.

        - Ano (`int`)
        - Janeiro (`float`)
        - Fevereiro (`float`)
        - ...
        - Dezembro (`float`)

        :return: As configurações em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        return self.__extrai_dados_se_existe(BlocoConfiguracoesExpansaoPMO, 0)

    @property
    def configuracoes_alteracao_potencia(self) -> Optional[pd.DataFrame]:
        """
        Configurações do sistema em cada período devido a alterações
        de potência.

        - Ano (`int`)
        - Janeiro (`float`)
        - Fevereiro (`float`)
        - ...
        - Dezembro (`float`)

        :return: As configurações em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        return self.__extrai_dados_se_existe(BlocoConfiguracoesExpansaoPMO, 1)

    @property
    def configuracoes_qualquer_modificacao(self) -> Optional[pd.DataFrame]:
        """
        Configurações do sistema em cada período devido a alterações
        de potência.

        - Ano (`int`)
        - Janeiro (`float`)
        - Fevereiro (`float`)
        - ...
        - Dezembro (`float`)

        :return: As configurações em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        return self.__extrai_dados_se_existe(BlocoConfiguracoesExpansaoPMO, 2)

    def retas_perdas_engolimento(self, estagio: int) -> Optional[pd.DataFrame]:
        """
        Retas ajustadas segundo o modelo MARS para corrigir a
        energia fio d'água com as perdas por engolimento máximo.

        - REE (`str`)
        - Reta (`int`)
        - Coeficiente Angular (`float`)
        - Constante (`float`)

        :return: As retas em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        return self.__extrai_dados_se_existe(BlocoMARSPMO, estagio - 1)

    @property
    def convergencia(self) -> Optional[pd.DataFrame]:
        """
        Tabela de convergência da execução do NEWAVE.

        - Iteração (`int`)
        - Lim. Inf. ZINF (`float`)
        - ZINF (`float`)
        - Lim. Sup. ZINF (`float`)
        - ZSUP (`float`)
        - Delta ZINF (`float`)
        - ZSUP Iteração (`float`)

        :return: As convergência em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        return self.__extrai_dados_se_existe(BlocoConvergenciaPMO)

    @property
    def risco_deficit_ens(self) -> Optional[pd.DataFrame]:
        """
        Tabela de riscos de déficit e enegia não suprida (ENS).

        - Ano (`int`)
        - Risco - <nome_subsistema_1> (`float`)
        - EENS - <nome_subsistema_1> (`float`)
        - ...
        - Risco - <nome_subsistema_N> (`float`)
        - EENS - <nome_subsistema_N> (`float`)

        :return: Os ricos em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        return self.__extrai_dados_se_existe(BlocoRiscoDeficitENSPMO)

    @property
    def custo_operacao_series_simuladas(self) -> Optional[pd.DataFrame]:
        """
        Tabela de custos de operação categorizados para as
        séries simuladas.

        - Parcela (`str`)
        - Valor Esperado (`float`)
        - Desvio Padrão do VE (`float`)
        - (%) (`float`)

        :return: Os custos em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        return self.__extrai_dados_se_existe(BlocoCustoOperacaoPMO, 0)

    @property
    def valor_esperado_periodo_estudo(self) -> Optional[pd.DataFrame]:
        """
        Tabela de custos de operação esperados para o período
        de estudo.

        - Parcela (`str`)
        - Valor Esperado (`float`)
        - Desvio Padrão do VE (`float`)
        - (%) (`float`)

        :return: Os custos em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        return self.__extrai_dados_se_existe(BlocoCustoOperacaoPMO, 1)

    @property
    def custo_operacao_referenciado_primeiro_mes(
        self,
    ) -> Optional[pd.DataFrame]:
        """
        Tabela de custos de operação esperados para o período
        de estudo, referenciados ao primeiro mês.

        - Parcela (`str`)
        - Valor Esperado (`float`)
        - Desvio Padrão do VE (`float`)
        - (%) (`float`)

        :return: Os custos em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        return self.__extrai_dados_se_existe(BlocoCustoOperacaoPMO, 2)

    @property
    def custo_operacao_total(self) -> Optional[float]:
        """
        Custo de operacao total da SF.

        :return: O custo total.
        :rtype: float | None
        """
        b = self.__extrai_dados_se_existe(BlocoCustoOperacaoTotalPMO)
        if isinstance(b, list):
            return b[0]
        return None

    @property
    def desvio_custo_operacao_total(self) -> Optional[float]:
        """
        O desvio padrão do custo de operacao total da SF.

        :return: O desvio do custo total.
        :rtype: float | None
        """
        b = self.__extrai_dados_se_existe(BlocoCustoOperacaoTotalPMO)
        if isinstance(b, list):
            return b[1]
        return None



# Rotinas de testes associadas ao arquivo vazpast.dat do NEWAVE
from inewave.newave.modelos.vazpast import BlocoVazPast

from inewave.newave import VazPast


from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.vazpast import MockBlocoVazoesPassadas


def test_bloco_desvios_vazpast():

    m: MagicMock = mock_open(read_data="".join(MockBlocoVazoesPassadas))
    b = BlocoVazPast()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 217
    assert b.data.iloc[0, 0] == 1
    assert b.data.iloc[-1, -1] == 20.0


def test_atributos_encontrados_vazpast():
    m: MagicMock = mock_open(read_data="".join(MockBlocoVazoesPassadas))
    with patch("builtins.open", m):
        ad = VazPast.le_arquivo("")
        assert ad.tendencia is not None


def test_atributos_nao_encontrados_vazpast():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        ad = VazPast.le_arquivo("")
        assert ad.tendencia is None


def test_eq_vazpast():
    m: MagicMock = mock_open(read_data="".join(MockBlocoVazoesPassadas))
    with patch("builtins.open", m):
        cf1 = VazPast.le_arquivo("")
        cf2 = VazPast.le_arquivo("")
        assert cf1 == cf2


def test_neq_vazpast():
    m: MagicMock = mock_open(read_data="".join(MockBlocoVazoesPassadas))
    with patch("builtins.open", m):
        cf1 = VazPast.le_arquivo("")
        cf2 = VazPast.le_arquivo("")
        cf2.tendencia.iloc[0, 0] = -1
        assert cf1 != cf2


def test_leitura_escrita_vazpast():
    m_leitura: MagicMock = mock_open(
        read_data="".join(MockBlocoVazoesPassadas)
    )
    with patch("builtins.open", m_leitura):
        cf1 = VazPast.le_arquivo("")
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        cf1.escreve_arquivo("", "")
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(2, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        cf2 = VazPast.le_arquivo("")
        assert cf1 == cf2



from inewave.newave.modelos.cvar import (
    BlocoValoresConstantesCVAR,
    BlocoAlfaVariavelNoTempo,
    BlocoLambdaVariavelNoTempo,
)

from cfinterface.files.blockfile import BlockFile
from typing import Type, TypeVar, Optional
import pandas as pd  # type: ignore


class CVAR(BlockFile):
    """
    Armazena os dados de entrada do NEWAVE referentes à curva para
    penalização por volume mínimo dos reservatórios.
    """

    T = TypeVar("T")

    BLOCKS = [
        BlocoValoresConstantesCVAR,
        BlocoAlfaVariavelNoTempo,
        BlocoLambdaVariavelNoTempo,
    ]

    def __init__(self, data=...) -> None:
        super().__init__(data)

    @classmethod
    def le_arquivo(cls, diretorio: str, nome_arquivo="cvar.dat") -> "CVAR":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="cvar.dat"):
        self.write(diretorio, nome_arquivo)

    def __bloco_por_tipo(self, bloco: Type[T], indice: int) -> Optional[T]:
        """
        Obtém um gerador de blocos de um tipo, se houver algum no arquivo.

        :param bloco: Um tipo de bloco para ser lido
        :type bloco: T
        :param indice: O índice do bloco a ser acessado, dentre os do tipo
        :type indice: int
        :return: O gerador de blocos, se houver
        :rtype: Optional[Generator[T], None, None]
        """
        try:
            return next(
                b
                for i, b in enumerate(self.data.of_type(bloco))
                if i == indice
            )
        except StopIteration:
            return None

    @property
    def valores_constantes(self) -> Optional[list]:
        """
        Valores constantes dos parâmetros ALFA e LAMBDA do CVAR.

        :return: Os valores dos campos da linha como uma lista.
        :rtype: list | None
        """
        b = self.__bloco_por_tipo(BlocoValoresConstantesCVAR, 0)
        if b is not None:
            return b.data
        return None

    @valores_constantes.setter
    def valores_constantes(self, valores: list):
        """
        Valores constantes dos parâmetros ALFA e LAMBDA do CVAR.

        :return: Os valores dos campos da linha como uma lista.
        :rtype: list | None
        """
        b = self.__bloco_por_tipo(BlocoValoresConstantesCVAR, 0)
        if b is not None:
            b.data = valores
        else:
            raise ValueError("Bloco não lido")

    @property
    def alfa_variavel(self) -> Optional[pd.DataFrame]:
        """
        Tabela com os valores variáveis do CVAR para o parâmetro ALFA.

        - Ano (`str`)
        - Janeiro (`float`)
        - Fevereiro (`float`)
        - ...
        - Dezembro (`float`)

        :return: O valor de ALFA por estágio em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.__bloco_por_tipo(BlocoAlfaVariavelNoTempo, 0)
        if b is not None:
            return b.data
        return None

    @property
    def lambda_variavel(self) -> Optional[pd.DataFrame]:
        """
        Tabela com os valores variáveis do CVAR para o parâmetro LAMBDA.

        - Ano (`str`)
        - Janeiro (`float`)
        - Fevereiro (`float`)
        - ...
        - Dezembro (`float`)

        :return: O valor de LAMBDA por estágio em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.__bloco_por_tipo(BlocoLambdaVariavelNoTempo, 0)
        if b is not None:
            return b.data
        return None
